from charms.layer import status
from charms.reactive import when, when_all, when_any, when_not, set_flag, clear_flag, hook
from charmhelpers.core.hookenv import charm_dir, config

from textwrap import dedent

import glob
import os
import shutil
import subprocess

AUTOPKGTEST_CLOUD_CONF = os.path.expanduser('~ubuntu/autopkgtest-cloud.conf')
GITHUB_SECRETS_PATH = os.path.expanduser('~ubuntu/github-secrets.json')
GITHUB_STATUS_CREDENTIALS_PATH = os.path.expanduser('~ubuntu/github-status-credentials.json')


@when_not('autopkgtest-web.autopkgtest_web_symlinked')
def symlink_autopkgtest_cloud():
    autopkgtest_cloud = os.path.join(charm_dir(), 'webcontrol')
    os.symlink(autopkgtest_cloud, os.path.expanduser('~ubuntu/webcontrol'))
    set_flag('autopkgtest-web.autopkgtest_web_symlinked')


@when('amqp.connected')
@when_not('amqp.available')
def setup_rabbitmq(rabbitmq):
    rabbitmq.request_access('webcontrol', '/')
    status.waiting('Waiting on RabbitMQ to configure vhost')


@when('amqp.available',
      'config.set.storage-url-internal')
def write_autopkgtest_cloud_conf(rabbitmq):
    swiftinternal = config().get('storage-url-internal')
    swiftexternal = config().get('storage-url-external') or swiftinternal
    rabbituser = rabbitmq.username()
    rabbitpassword = rabbitmq.password()
    rabbithost = rabbitmq.private_address()
    clear_flag('autopkgtest-web.config-written')
    with open(AUTOPKGTEST_CLOUD_CONF, 'w') as f:
        f.write(dedent('''\
            [web]
            database=/home/ubuntu/autopkgtest.db
            database_ro=/home/ubuntu/public/autopkgtest.db
            SwiftURL={swiftinternal}
            SwiftURLExternal={swiftexternal}

            [amqp]
            uri=amqp://{rabbituser}:{rabbitpassword}@{rabbithost}'''.format(**locals())))
    set_flag('autopkgtest-web.config-written')


@when('autopkgtest-web.autopkgtest_web_symlinked',
      'autopkgtest-web.config-written')
def set_up_systemd_units():
    status.maintenance('Installing systemd units')
    for unit in glob.glob(os.path.join(charm_dir(), 'units', '*')):
        base = os.path.basename(unit)
        try:
            os.symlink(unit, os.path.join(os.path.sep,
                                          'etc',
                                          'systemd',
                                          'system',
                                          base))
            subprocess.check_call(['systemctl',
                                   'daemon-reload'])
            subprocess.check_call(['systemctl',
                                   'enable',
                                   base])
        except FileExistsError:
            pass

    subprocess.check_call(['systemctl',
                           'restart',
                           'autopkgtest-web.target'])

    status.active('systemd units installed and enabled')


@when_any('config.set.https-proxy',
          'config.set.no-proxy',
          'website.available')
@when('website.available')
def set_up_web_config(apache):
    webcontrol_dir = os.path.join(charm_dir(), 'webcontrol')
    sn = config().get('hostname')
    https_proxy = config().get('https-proxy')
    no_proxy = config().get('no-proxy')

    conf_enabled = os.path.join(os.path.sep, 'etc', 'apache2', 'conf-enabled')
    environment_d = os.path.join(os.path.sep, 'etc', 'environment.d')

    if https_proxy or no_proxy:
        try:
            os.makedirs(environment_d)
        except FileExistsError:
            pass

    https_proxy_env = os.path.join(environment_d, 'https_proxy.conf')
    https_proxy_apache = os.path.join(conf_enabled, 'https_proxy.conf')

    try:
        os.unlink(https_proxy_env)
    except FileNotFoundError:
        pass
    try:
        os.unlink(https_proxy_apache)
    except FileNotFoundError:
        pass

    if https_proxy:
        with open(https_proxy_env, 'w') as f:
            f.write('https_proxy={}'.format(https_proxy))
        with open(https_proxy_apache, 'w') as f:
            f.write('SetEnv https_proxy {}'.format(https_proxy))

    no_proxy_env = os.path.join(environment_d, 'no_proxy.conf')
    no_proxy_apache = os.path.join(conf_enabled, 'no_proxy.conf')

    try:
        os.unlink(no_proxy_env)
    except FileNotFoundError:
        pass
    try:
        os.unlink(no_proxy_apache)
    except FileNotFoundError:
        pass

    if no_proxy:
        with open(no_proxy_env, 'w') as f:
            f.write('no_proxy={}'.format(no_proxy))
        with open(no_proxy_apache, 'w') as f:
            f.write('SetEnv no_proxy {}'.format(no_proxy))

    server_name = 'ServerName {}'.format(sn) if sn else ''
    apache.send_site_config(dedent('''\
        <Directory {webcontrol_dir}>
                Require all granted
        </Directory>

        <VirtualHost *:80>
                DocumentRoot {webcontrol_dir}
                ErrorLog ${{APACHE_LOG_DIR}}/error.log
                CustomLog ${{APACHE_LOG_DIR}}/access.log combined
                AddType 'text/plain' .log
                AddCharset UTF-8 .log
                DirectoryIndex browse.cgi
                RedirectMatch permanent "^/packages(.*)/$" "/packages$1"
                Redirect permanent /running.shtml /running

                # webcontrol CGI scripts
                ScriptAlias /request.cgi {webcontrol_dir}/request.cgi/
                ScriptAlias /login {webcontrol_dir}/request.cgi/login
                ScriptAlias /logout {webcontrol_dir}/request.cgi/logout
                ScriptAlias / {webcontrol_dir}/browse.cgi/
                {server_name}
        </VirtualHost>'''.format(webcontrol_dir=webcontrol_dir,
                                 server_name=server_name)))
    apache.send_ports([80])  # XXX: 443
    apache.send_enabled()


@when_all('config.changed.github-secrets',
          'config.set.github-secrets')
def write_github_secrets():
    github_secrets = config().get('github-secrets')

    with open(GITHUB_SECRETS_PATH, 'w') as f:
        f.write(github_secrets)

    try:
        os.symlink(GITHUB_SECRETS_PATH,
                   os.path.expanduser('~www-data/github-secrets.json'))
    except FileExistsError:
        pass


@when_not('config.set.github-secrets')
def clear_github_secrets():
    try:
        os.unlink(GITHUB_SECRETS_PATH)
    except FileNotFoundError:
        pass

    try:
        os.unlink(os.path.expanduser('~www-data/github-secrets.json'))
    except FileNotFoundError:
        pass


@when_all('config.changed.github-status-credentials',
          'config.set.github-status-credentials')
def write_github_status_credentials():
    github_status_credentials = config().get('github-status-credentials')

    with open(GITHUB_STATUS_CREDENTIALS_PATH, 'w') as f:
        f.write(github_status_credentials)

    try:
        os.symlink(GITHUB_STATUS_CREDENTIALS_PATH,
                   os.path.expanduser('~www-data/github-status-credentials.json'))
    except FileExistsError:
        pass


@when_not('config.set.github-status-credentials')
def clear_github_status_credentials():
    try:
        os.unlink(GITHUB_STATUS_CREDENTIALS_PATH)
    except FileNotFoundError:
        pass

    try:
        os.unlink(os.path.expanduser('~www-data/github-status-credentials.json'))
    except FileNotFoundError:
        pass


@when_not('autopkgtest-web.bootstrap-symlinked')
def symlink_bootstrap():
    try:
        os.symlink(os.path.join(os.path.sep,
                                'usr',
                                'share',
                                'javascript',
                                'bootstrap'),
                   os.path.join(charm_dir(),
                                'webcontrol',
                                'static',
                                'bootstrap'))
        set_flag('autopkgtest-web.bootstrap-symlinked')
    except FileExistsError:
        pass


@when_not('autopkgtest-web.runtime-dir-created')
def make_runtime_tmpfiles():
    with open('/etc/tmpfiles.d/autopkgtest-web-runtime.conf', 'w') as r:
        r.write('D %t/autopkgtest_webcontrol 0755 www-data www-data')
    set_flag('autopkgtest-web.runtime-dir-created')


@when_not('autopkgtest-web.running-json-symlinked')
def symlink_running():
    try:
        os.symlink(os.path.join(os.path.sep,
                                'run',
                                'amqp-status-collector',
                                'running.json'),
                   os.path.join(charm_dir(),
                                'webcontrol',
                                'static',
                                'running.json'))
        set_flag('autopkgtest-web.running-json-symlinked')
    except FileExistsError:
        pass


@when_not('autopkgtest-web.public-db-symlinked')
def symlink_public_db():
    try:
        publicdir = os.path.expanduser('~ubuntu/public/')
        os.makedirs(publicdir)
        shutil.chown(publicdir, user='ubuntu', group='ubuntu')
        os.symlink(os.path.join(publicdir,
                                'autopkgtest.db'),
                   os.path.join(charm_dir(),
                                'webcontrol',
                                'static',
                                'autopkgtest.db'))
        set_flag('autopkgtest-web.public-db-symlinked')
    except FileExistsError:
        pass
